@@ -26,14 +26,21 @@ import google.api_core.client_options
 from google.auth.credentials import AnonymousCredentials
 
 from google.api_core import page_iterator
-from google.cloud._helpers import _LocalStack, _NOW
+from google.cloud._helpers import _LocalStack
 from google.cloud.client import ClientWithProject
 from google.cloud.exceptions import NotFound
 
-from google.cloud.storage._helpers import _get_environ_project
-from google.cloud.storage._helpers import _get_storage_host
-from google.cloud.storage._helpers import _DEFAULT_STORAGE_HOST
 from google.cloud.storage._helpers import _bucket_bound_hostname_url
+from google.cloud.storage._helpers import _get_api_endpoint_override
+from google.cloud.storage._helpers import _get_environ_project
+from google.cloud.storage._helpers import _get_storage_emulator_override
+from google.cloud.storage._helpers import _use_client_cert
+from google.cloud.storage._helpers import _virtual_hosted_style_base_url
+from google.cloud.storage._helpers import _DEFAULT_UNIVERSE_DOMAIN
+from google.cloud.storage._helpers import _DEFAULT_SCHEME
+from google.cloud.storage._helpers import _STORAGE_HOST_TEMPLATE
+from google.cloud.storage._helpers import _NOW
+from google.cloud.storage._helpers import _UTC
 
 from google.cloud.storage._http import Connection
 from google.cloud.storage._signing import (
@@ -87,13 +94,18 @@ class Client(ClientWithProject):
 
     :type client_options: :class:`~google.api_core.client_options.ClientOptions` or :class:`dict`
     :param client_options: (Optional) Client options used to set user options on the client.
-        API Endpoint should be set through client_options.
+        A non-default universe domain or api endpoint should be set through client_options.
 
     :type use_auth_w_custom_endpoint: bool
     :param use_auth_w_custom_endpoint:
         (Optional) Whether authentication is required under custom endpoints.
         If false, uses AnonymousCredentials and bypasses authentication.
         Defaults to True. Note this is only used when a custom endpoint is set in conjunction.
+
+    :type extra_headers: dict
+    :param extra_headers:
+        (Optional) Custom headers to be sent with the requests attached to the client.
+        For example, you can add custom audit logging headers.
     """
 
     SCOPE = (
@@ -111,6 +123,7 @@ class Client(ClientWithProject):
         client_info=None,
         client_options=None,
         use_auth_w_custom_endpoint=True,
+        extra_headers={},
     ):
         self._base_connection = None
 
@@ -127,33 +140,81 @@ class Client(ClientWithProject):
         # are passed along, for use in __reduce__ defined elsewhere.
         self._initial_client_info = client_info
         self._initial_client_options = client_options
+        self._extra_headers = extra_headers
 
-        kw_args = {"client_info": client_info}
-
-        # `api_endpoint` should be only set by the user via `client_options`,
-        # or if the _get_storage_host() returns a non-default value (_is_emulator_set).
-        # `api_endpoint` plays an important role for mTLS, if it is not set,
-        # then mTLS logic will be applied to decide which endpoint will be used.
-        storage_host = _get_storage_host()
-        _is_emulator_set = storage_host != _DEFAULT_STORAGE_HOST
-        kw_args["api_endpoint"] = storage_host if _is_emulator_set else None
+        connection_kw_args = {"client_info": client_info}
 
         if client_options:
             if isinstance(client_options, dict):
                 client_options = google.api_core.client_options.from_dict(
                     client_options
                 )
-            if client_options.api_endpoint:
-                api_endpoint = client_options.api_endpoint
-                kw_args["api_endpoint"] = api_endpoint
+
+        if client_options and client_options.universe_domain:
+            self._universe_domain = client_options.universe_domain
+        else:
+            self._universe_domain = None
+
+        storage_emulator_override = _get_storage_emulator_override()
+        api_endpoint_override = _get_api_endpoint_override()
+
+        # Determine the api endpoint. The rules are as follows:
+
+        # 1. If the `api_endpoint` is set in `client_options`, use that as the
+        #    endpoint.
+        if client_options and client_options.api_endpoint:
+            api_endpoint = client_options.api_endpoint
+
+        # 2. Elif the "STORAGE_EMULATOR_HOST" env var is set, then use that as the
+        #    endpoint.
+        elif storage_emulator_override:
+            api_endpoint = storage_emulator_override
+
+        # 3. Elif the "API_ENDPOINT_OVERRIDE" env var is set, then use that as the
+        #    endpoint.
+        elif api_endpoint_override:
+            api_endpoint = api_endpoint_override
+
+        # 4. Elif the `universe_domain` is set in `client_options`,
+        #    create the endpoint using that as the default.
+        #
+        #    Mutual TLS is not compatible with a non-default universe domain
+        #    at this time. If such settings are enabled along with the
+        #    "GOOGLE_API_USE_CLIENT_CERTIFICATE" env variable, a ValueError will
+        #    be raised.
+
+        elif self._universe_domain:
+            # The final decision of whether to use mTLS takes place in
+            # google-auth-library-python. We peek at the environment variable
+            # here only to issue an exception in case of a conflict.
+            if _use_client_cert():
+                raise ValueError(
+                    'The "GOOGLE_API_USE_CLIENT_CERTIFICATE" env variable is '
+                    'set to "true" and a non-default universe domain is '
+                    "configured. mTLS is not supported in any universe other than"
+                    "googleapis.com."
+                )
+            api_endpoint = _DEFAULT_SCHEME + _STORAGE_HOST_TEMPLATE.format(
+                universe_domain=self._universe_domain
+            )
+
+        # 5. Else, use the default, which is to use the default
+        #    universe domain of "googleapis.com" and create the endpoint
+        #    "storage.googleapis.com" from that.
+        else:
+            api_endpoint = None
+
+        connection_kw_args["api_endpoint"] = api_endpoint
+
+        self._is_emulator_set = True if storage_emulator_override else False
 
         # If a custom endpoint is set, the client checks for credentials
         # or finds the default credentials based on the current environment.
         # Authentication may be bypassed under certain conditions:
         # (1) STORAGE_EMULATOR_HOST is set (for backwards compatibility), OR
         # (2) use_auth_w_custom_endpoint is set to False.
-        if kw_args["api_endpoint"] is not None:
-            if _is_emulator_set or not use_auth_w_custom_endpoint:
+        if connection_kw_args["api_endpoint"] is not None:
+            if self._is_emulator_set or not use_auth_w_custom_endpoint:
                 if credentials is None:
                     credentials = AnonymousCredentials()
                 if project is None:
@@ -169,10 +230,26 @@ class Client(ClientWithProject):
             _http=_http,
         )
 
+        # Validate that the universe domain of the credentials matches the
+        # universe domain of the client.
+        if self._credentials.universe_domain != self.universe_domain:
+            raise ValueError(
+                "The configured universe domain ({client_ud}) does not match "
+                "the universe domain found in the credentials ({cred_ud}). If "
+                "you haven't configured the universe domain explicitly, "
+                "`googleapis.com` is the default.".format(
+                    client_ud=self.universe_domain,
+                    cred_ud=self._credentials.universe_domain,
+                )
+            )
+
         if no_project:
             self.project = None
 
-        self._connection = Connection(self, **kw_args)
+        # Pass extra_headers to Connection
+        connection = Connection(self, **connection_kw_args)
+        connection.extra_headers = extra_headers
+        self._connection = connection
         self._batch_stack = _LocalStack()
 
     @classmethod
@@ -191,6 +268,14 @@ class Client(ClientWithProject):
         client.project = None
         return client
 
+    @property
+    def universe_domain(self):
+        return self._universe_domain or _DEFAULT_UNIVERSE_DOMAIN
+
+    @property
+    def api_endpoint(self):
+        return self._connection.API_BASE_URL
+
     @property
     def _connection(self):
         """Get connection or batch on the client.
@@ -835,6 +920,7 @@ class Client(ClientWithProject):
         data_locations=None,
         predefined_acl=None,
         predefined_default_object_acl=None,
+        enable_object_retention=False,
         timeout=_DEFAULT_TIMEOUT,
         retry=DEFAULT_RETRY,
     ):
@@ -873,6 +959,9 @@ class Client(ClientWithProject):
             predefined_default_object_acl (str):
                 (Optional) Name of predefined ACL to apply to bucket's objects. See:
                 https://cloud.google.com/storage/docs/access-control/lists#predefined-acl
+            enable_object_retention (bool):
+                (Optional) Whether object retention should be enabled on this bucket. See:
+                https://cloud.google.com/storage/docs/object-lock
             timeout (Optional[Union[float, Tuple[float, float]]]):
                 The amount of time, in seconds, to wait for the server response.
 
@@ -908,8 +997,7 @@ class Client(ClientWithProject):
             project = self.project
 
         # Use no project if STORAGE_EMULATOR_HOST is set
-        _is_emulator_set = _get_storage_host() != _DEFAULT_STORAGE_HOST
-        if _is_emulator_set:
+        if self._is_emulator_set:
             if project is None:
                 project = _get_environ_project()
             if project is None:
@@ -941,6 +1029,9 @@ class Client(ClientWithProject):
         if user_project is not None:
             query_params["userProject"] = user_project
 
+        if enable_object_retention:
+            query_params["enableObjectRetention"] = enable_object_retention
+
         properties = {key: bucket._properties[key] for key in bucket._changes}
         properties["name"] = bucket.name
 
@@ -1321,8 +1412,7 @@ class Client(ClientWithProject):
             project = self.project
 
         # Use no project if STORAGE_EMULATOR_HOST is set
-        _is_emulator_set = _get_storage_host() != _DEFAULT_STORAGE_HOST
-        if _is_emulator_set:
+        if self._is_emulator_set:
             if project is None:
                 project = _get_environ_project()
             if project is None:
@@ -1557,13 +1647,16 @@ class Client(ClientWithProject):
                             key to sign text.
 
         :type virtual_hosted_style: bool
-        :param virtual_hosted_style: (Optional) If True, construct the URL relative to the bucket
-                                     virtual hostname, e.g., '<bucket-name>.storage.googleapis.com'.
+        :param virtual_hosted_style:
+            (Optional) If True, construct the URL relative to the bucket
+            virtual hostname, e.g., '<bucket-name>.storage.googleapis.com'.
+            Incompatible with bucket_bound_hostname.
 
         :type bucket_bound_hostname: str
         :param bucket_bound_hostname:
             (Optional) If passed, construct the URL relative to the bucket-bound hostname.
             Value can be bare or with a scheme, e.g., 'example.com' or 'http://example.com'.
+            Incompatible with virtual_hosted_style.
             See: https://cloud.google.com/storage/docs/request-endpoints#cname
 
         :type scheme: str
@@ -1578,9 +1671,17 @@ class Client(ClientWithProject):
         :type access_token: str
         :param access_token: (Optional) Access token for a service account.
 
+        :raises: :exc:`ValueError` when mutually exclusive arguments are used.
+
         :rtype: dict
         :returns: Signed POST policy.
         """
+        if virtual_hosted_style and bucket_bound_hostname:
+            raise ValueError(
+                "Only one of virtual_hosted_style and bucket_bound_hostname "
+                "can be specified."
+            )
+
         credentials = self._credentials if credentials is None else credentials
         ensure_signed_credentials(credentials)
 
@@ -1608,7 +1709,7 @@ class Client(ClientWithProject):
         conditions += required_conditions
 
         # calculate policy expiration time
-        now = _NOW()
+        now = _NOW(_UTC).replace(tzinfo=None)
         if expiration is None:
             expiration = now + datetime.timedelta(hours=1)
 
@@ -1652,11 +1753,13 @@ class Client(ClientWithProject):
         )
         # designate URL
         if virtual_hosted_style:
-            url = f"https://{bucket_name}.storage.googleapis.com/"
+            url = _virtual_hosted_style_base_url(
+                self.api_endpoint, bucket_name, trailing_slash=True
+            )
         elif bucket_bound_hostname:
             url = f"{_bucket_bound_hostname_url(bucket_bound_hostname, scheme)}/"
         else:
-            url = f"https://storage.googleapis.com/{bucket_name}/"
+            url = f"{self.api_endpoint}/{bucket_name}/"
 
         return {"url": url, "fields": policy_fields}
 
